"""'mediatool.py'."""

import pandas as pd
from prefect import get_run_logger, task

from viadot.orchestration.prefect.exceptions import MissingSourceCredentialsError
from viadot.orchestration.prefect.utils import get_credentials
from viadot.sources import Mediatool
from viadot.utils import join_dfs


@task(retries=3, log_prints=True, retry_delay_seconds=10, timeout_seconds=60 * 60)
def mediatool_to_df(
    config_key: str | None = None,
    azure_key_vault_secret: str | None = None,
    organization_ids: list[str] | None = None,
    media_entries_columns: list[str] | None = None,
) -> pd.DataFrame:
    """Task to download data from Mediatool API.

    Data from different endpoints are retrieved and combined. A final object is created
    containing data for all organizations from the list.

    Args:
        config_key (str, optional): The key in the viadot config holding relevant
            credentials. Defaults to None.
        azure_key_vault_secret (str, optional): The name of the Azure Key Vault secret
            where credentials are stored. Defaults to None.
        organization_ids (list[str], optional): List of organization IDs.
            Defaults to None.
        media_entries_columns (list[str], optional): Columns to get from media entries.
            Defaults to None.

    Raises:
        ValueError: 'organization_ids' argument is None, and one is mandatory.
        ValueError: One 'organization_id' is not in organizations.

    Returns:
        pd.DataFrame: The response data as a Pandas Data Frame.
    """
    logger = get_run_logger()

    if not (azure_key_vault_secret or config_key):
        raise MissingSourceCredentialsError

    if not config_key:
        credentials = get_credentials(azure_key_vault_secret)

    mediatool = Mediatool(
        credentials=credentials,
        config_key=config_key,
    )
    # first method ORGANIZATIONS
    method = "organizations"
    organizations_data = mediatool.api_connection(get_data_from=method)
    df_organizations = mediatool.to_df(data=organizations_data, column_suffix=method)

    if organization_ids is None:
        message = "No organizations were defined."
        raise ValueError(message)

    list_of_organizations_df = []
    for organization_id in organization_ids:
        if organization_id in df_organizations["_id_organizations"].unique():
            logger.info(f"Downloading data for: {organization_id} ...")

            # extract media entries per organization
            media_entries_data = mediatool.api_connection(
                get_data_from="media_entries",
                organization_id=organization_id,
            )
            df_media_entries = mediatool.to_df(
                data=media_entries_data, drop_columns=media_entries_columns
            )
            unique_vehicle_ids = df_media_entries["vehicleId"].unique()
            unique_media_type_ids = df_media_entries["mediaTypeId"].unique()

            # extract vehicles
            method = "vehicles"
            vehicles_data = mediatool.api_connection(
                get_data_from=method,
                vehicle_ids=unique_vehicle_ids,
            )
            df_vehicles = mediatool.to_df(data=vehicles_data, column_suffix=method)

            # extract campaigns
            method = "campaigns"
            campaigns_data = mediatool.api_connection(
                get_data_from=method,
                organization_id=organization_id,
            )
            df_campaigns = mediatool.to_df(data=campaigns_data, column_suffix=method)

            # extract media types
            method = "media_types"
            media_types_data = mediatool.api_connection(
                get_data_from=method,
                media_type_ids=unique_media_type_ids,
            )
            df_media_types = mediatool.to_df(
                data=media_types_data, column_suffix=method
            )

            # join media entries & organizations
            df_merged_entries_orgs = join_dfs(
                df_left=df_media_entries,
                df_right=df_organizations,
                left_on="organizationId",
                right_on="_id_organizations",
                columns_from_right_df=[
                    "_id_organizations",
                    "name_organizations",
                    "abbreviation_organizations",
                ],
                how="left",
            )

            # join the previous merge & campaigns
            df_merged_campaigns = join_dfs(
                df_left=df_merged_entries_orgs,
                df_right=df_campaigns,
                left_on="campaignId",
                right_on="_id_campaigns",
                columns_from_right_df=[
                    "_id_campaigns",
                    "name_campaigns",
                    "conventionalName_campaigns",
                ],
                how="left",
            )

            # join the previous merge & vehicles
            df_merged_vehicles = join_dfs(
                df_left=df_merged_campaigns,
                df_right=df_vehicles,
                left_on="vehicleId",
                right_on="_id_vehicles",
                columns_from_right_df=["_id_vehicles", "name_vehicles"],
                how="left",
            )

            # join the previous merge & media types
            df_merged_media_types = join_dfs(
                df_left=df_merged_vehicles,
                df_right=df_media_types,
                left_on="mediaTypeId",
                right_on="_id_media_types",
                columns_from_right_df=["_id_media_types", "name_media_types"],
                how="left",
            )

            list_of_organizations_df.append(df_merged_media_types)

        else:
            message = (
                f"Organization - {organization_id} not found in organizations list."
            )
            raise ValueError(message)

    return pd.concat(list_of_organizations_df)



"""Sharepoint API connector."""

import io
from pathlib import Path
import re
from typing import Any, Literal
from urllib.parse import urlparse

import pandas as pd
from pandas._libs.parsers import STR_NA_VALUES
from pydantic import BaseModel, root_validator
import sharepy
from sharepy.errors import AuthError

from viadot.config import get_source_credentials
from viadot.exceptions import CredentialError
from viadot.signals import SKIP
from viadot.sources.base import Source
from viadot.utils import (
    add_viadot_metadata_columns,
    cleanup_df,
    validate,
    validate_and_reorder_dfs_columns,
)


class SharepointCredentials(BaseModel):
    site: str  # Path to sharepoint website (e.g : {tenant_name}.sharepoint.com)
    username: str  # Sharepoint username (e.g username@{tenant_name}.com)
    password: str  # Sharepoint password

    @root_validator(pre=True)
    def is_configured(cls, credentials: dict):  # noqa: N805, ANN201, D102
        site = credentials.get("site")
        username = credentials.get("username")
        password = credentials.get("password")

        if not (site and username and password):
            msg = "'site', 'username', and 'password' credentials are required."
            raise CredentialError(msg)
        return credentials


class Sharepoint(Source):
    DEFAULT_NA_VALUES = tuple(STR_NA_VALUES)

    def __init__(
        self,
        credentials: SharepointCredentials = None,
        config_key: str | None = None,
        *args,
        **kwargs,
    ):
        """Download Excel files from Sharepoint.

        Args:
        credentials (SharepointCredentials): Sharepoint credentials.
        config_key (str, optional): The key in the viadot config holding relevant
            credentials.
        """
        raw_creds = credentials or get_source_credentials(config_key) or {}
        validated_creds = dict(SharepointCredentials(**raw_creds))
        super().__init__(*args, credentials=validated_creds, **kwargs)

    def get_connection(self) -> sharepy.session.SharePointSession:
        """Establishe a connection to SharePoint.

        Returns:
            sharepy.session.SharePointSession: A session object representing
                the authenticated connection.

        Raises:
            CredentialError: If authentication to SharePoint fails due to incorrect
                credentials.
        """
        try:
            connection = sharepy.connect(
                site=self.credentials.get("site"),
                username=self.credentials.get("username"),
                password=self.credentials.get("password"),
            )
        except AuthError as e:
            site = self.credentials.get("site")
            msg = f"Could not authenticate to {site} with provided credentials."
            raise CredentialError(msg) from e
        return connection

    def download_file(self, url: str, to_path: list | str) -> None:
        """Download a file from Sharepoint to specific location.

        Args:
            url (str): The URL of the file to be downloaded.
            to_path (str): Where to download the file.

        Example:
            download_file(
                url="https://{tenant_name}.sharepoint.com/sites/{directory}/Shared%20Documents/Dashboard/file",
                to_path="file.xlsx"
            )
        """
        conn = self.get_connection()
        conn.getfile(
            url=url,
            filename=to_path,
        )
        conn.close()

    def scan_sharepoint_folder(self, url: str) -> list[str]:
        """Scan Sharepoint folder to get all file URLs of all files within it.

        Args:
            url (str): The URL of the folder to scan.

        Raises:
            ValueError: If the provided URL does not contain the expected '/sites/'
                segment.

        Returns:
            list[str]: List of URLs pointing to each file within the specified
                SharePoint folder.
        """
        conn = self.get_connection()

        parsed_url = urlparse(url)
        path_parts = parsed_url.path.split("/")
        if "sites" in path_parts:
            site_index = (
                path_parts.index("sites") + 2
            )  # +2 to include 'sites' and the next segment
            site_url = f"{parsed_url.scheme}://{parsed_url.netloc}{'/'.join(path_parts[:site_index])}"
            library = "/".join(path_parts[site_index:])
        else:
            message = "URL does not contain '/sites/' segment."
            raise ValueError(message)

        # -> site_url = company.sharepoint.com/sites/site_name/
        # -> library = /shared_documents/folder/sub_folder/final_folder
        endpoint = (
            f"{site_url}/_api/web/GetFolderByServerRelativeUrl('{library}')/Files"
        )
        response = conn.get(endpoint)
        files = response.json().get("d", {}).get("results", [])

        return [f'{site_url}/{library}{file["Name"]}' for file in files]

    def _get_file_extension(self, url: str) -> str:
        """Extracts the file extension from a given URL.

        Parameters:
            url (str): The URL from which to extract the file extension.

        Returns:
            str: The file extension, including the leading dot (e.g., '.xlsx').
        """
        # Parse the URL to get the path
        parsed_url = urlparse(url)
        return Path(parsed_url.path).suffix

    def _download_file_stream(self, url: str, **kwargs) -> pd.ExcelFile:
        """Download the contents of a file from SharePoint.

        Returns the data as an in-memory byte stream.

        Args:
            url (str): The URL of the file to download.

        Returns:
            io.BytesIO: An in-memory byte stream containing the file content.
        """
        if "nrows" in kwargs:
            msg = "Parameter 'nrows' is not supported."
            raise ValueError(msg)

        conn = self.get_connection()

        self.logger.info(f"Downloading data from {url}...")
        response = conn.get(url)
        bytes_stream = io.BytesIO(response.content)

        return pd.ExcelFile(bytes_stream)

    def _is_file(self, url: str) -> bool:
        """Determines whether a provided URL points to a file based on its structure.

        This function uses a regular expression to check if the URL ends with a
        common file extension. It does not make any network requests and purely
        relies on the URL structure for its determination.

        Parameters:
        url (str): The URL to be checked.

        Returns:
        bool: True if the URL is identified as a file based on its extension,
            False otherwise.

        Example:
        >>> _is_file("https://example.com/file.xlsx")
        True
        >>> _is_file("https://example.com/folder/")
        False
        >>> _is_file("https://example.com/folder")
        False
        """
        # Regular expression for matching file extensions
        file_extension_pattern = re.compile(r"\.[a-zA-Z0-9]+$")

        return bool(file_extension_pattern.search(url))

    def _handle_multiple_files(
        self,
        url: str,
        file_sheet_mapping: dict,
        na_values: list[str] | None = None,
        **kwargs,
    ):
        """Handle downloading and parsing multiple Excel files from a SharePoint folder.

        Args:
            url (str): The base URL of the SharePoint folder containing the files.
            file_sheet_mapping (dict): A dictionary mapping file names to sheet names
                or indexes. The keys are file names, and the values are sheet
                names/indices.
            na_values (Optional[list[str]]): Additional strings to recognize as NA/NaN.

        Returns:
            pd.DataFrame: A concatenated DataFrame containing the data from all
                specified files and sheets.

        Raises:
            ValueError: If the file extension is not supported.
        """
        dfs = [
            self._load_and_parse(
                file_url=url + file, sheet_name=sheet, na_values=na_values, **kwargs
            )
            for file, sheet in file_sheet_mapping.items()
        ]
        return pd.concat(validate_and_reorder_dfs_columns(dfs))

    def _load_and_parse(
        self,
        file_url: str,
        sheet_name: str | list[str] | None = None,
        na_values: list[str] | None = None,
        **kwargs,
    ):
        """Loads and parses an Excel file from a URL.

        Args:
            file_url (str): The URL of the file to download and parse.
            sheet_name (Optional[Union[str, list[str]]]): The name(s) or index(es) of
                the sheet(s) to parse. If None, all sheets are parsed.
            na_values (Optional[list[str]]): Additional strings to recognize as NA/NaN.
            **kwargs: Additional keyword arguments to pass to the pandas read function.

        Returns:
            pd.DataFrame: The parsed data as a pandas DataFrame.

        Raises:
            ValueError: If the file extension is not supported.
        """
        file_extension = self._get_file_extension(file_url)
        file_stream = self._download_file_stream(file_url)

        if file_extension == ".xlsx":
            return self._parse_excel(file_stream, sheet_name, na_values, **kwargs)
        msg = "Only Excel (.xlsx) files can be loaded into a DataFrame."
        raise ValueError(msg)

    def _parse_excel(
        self,
        excel_file: pd.ExcelFile,
        sheet_name: str | list[str] | None = None,
        na_values: list[str] | None = None,
        **kwargs,
    ):
        """Parses an Excel file into a DataFrame. Casts all columns to string.

        Args:
            excel_file: An ExcelFile object containing the data to parse.
            sheet_name (Optional[Union[str, list[str]]]): The name(s) or index(es) of
                the sheet(s) to parse. If None, all sheets are parsed.
            na_values (Optional[list[str]]): Additional strings to recognize as NA/NaN.
            **kwargs: Additional keyword arguments to pass to the pandas read function.

        Returns:
            pd.DataFrame: The parsed data as a pandas DataFrame.
        """
        return pd.concat(
            [
                excel_file.parse(
                    sheet,
                    keep_default_na=False,
                    na_values=na_values or list(self.DEFAULT_NA_VALUES),
                    dtype=str,  # Ensure all columns are read as strings
                    **kwargs,
                )
                for sheet in ([sheet_name] if sheet_name else excel_file.sheet_names)
            ]
        )

    @add_viadot_metadata_columns
    def to_df(
        self,
        url: str,
        sheet_name: str | list[str] | None = None,
        if_empty: Literal["warn", "skip", "fail"] = "warn",
        tests: dict[str, Any] | None = None,
        file_sheet_mapping: dict[str, str | int | list[str]] | None = None,
        na_values: list[str] | None = None,
        **kwargs,
    ) -> pd.DataFrame:
        """Load an Excel file or files from a SharePoint URL into a pandas DataFrame.

        This method handles downloading the file(s), parsing the content, and converting
        it into a pandas DataFrame. It supports both single file URLs and folder URLs
        with multiple files.

        Args:
            url (str): The URL of the file to be downloaded.
            sheet_name (Optional[Union[str, list, int]], optional): Strings are used for
                sheet names. Integers are used in zero-indexed sheet positions (chart
                sheets do not count as a sheet position). Lists of strings/integers are
                used to request multiple sheets. Specify None to get all worksheets.
                Defaults to None.
            if_empty (Literal["warn", "skip", "fail"], optional): Action to take if
                the DataFrame is empty.
                - "warn": Logs a warning.
                - "skip": Skips the operation.
                - "fail": Raises an error.
                Defaults to "warn".
            tests (Dict[str, Any], optional): A dictionary with optional list of tests
                to verify the output dataframe. If defined, triggers the `validate`
                function from utils. Defaults to None.
            file_sheet_mapping (dict[str, Union[str, int, list[str]]], optional):
                Mapping of file names to sheet names or indices. The keys are file names
                and the values are sheet names/indices. Used when multiple files are
                involved. Defaults to None.
            na_values (list[str], optional): Additional strings to recognize as NA/NaN.
                If list passed, the specific NA values for each column will be
                recognized. Defaults to None.
            kwargs (dict[str, Any], optional): Keyword arguments to pass to
                pd.ExcelFile.parse(). Note that `nrows` is not supported.

        Returns:
            pd.DataFrame: The resulting data as a pandas DataFrame.

        Raises:
            ValueError: If the file extension is not supported or if `if_empty` is set
                to "fail" and the DataFrame is empty.
            SKIP: If `if_empty` is set to "skip" and the DataFrame is empty.
        """
        if self._is_file(url):
            df = self._load_and_parse(
                file_url=url, sheet_name=sheet_name, na_values=na_values, **kwargs
            )
        elif file_sheet_mapping:
            df = self._handle_multiple_files(
                url=url,
                file_sheet_mapping=file_sheet_mapping,
                na_values=na_values,
                **kwargs,
            )
        else:
            list_of_urls = self.scan_sharepoint_folder(url)
            dfs = [
                self._load_and_parse(
                    file_url=file_url,
                    sheet_name=sheet_name,
                    na_values=na_values,
                    **kwargs,
                )
                for file_url in list_of_urls
            ]
            df = pd.concat(validate_and_reorder_dfs_columns(dfs))

        if df.empty:
            try:
                self._handle_if_empty(if_empty)
            except SKIP:
                return pd.DataFrame()
        else:
            self.logger.info(f"Successfully downloaded {len(df)} rows of data.")

        df_clean = cleanup_df(df)

        if tests:
            validate(df=df_clean, tests=tests)

        return df_clean



"""Build specified dbt model(s) and upload the generated metadata to Luma."""

from pathlib import Path
import shutil
from typing import Literal

from prefect import allow_failure, flow, task

from viadot.orchestration.prefect.tasks import clone_repo, dbt_task, luma_ingest_task
from viadot.orchestration.prefect.utils import get_credentials


@task
def remove_dbt_repo_dir(dbt_repo_dir_name: str) -> None:
    """Remove the repo directory.

    Args:
        dbt_repo_dir_name (str): The name of the dbt repo directory.
    """
    shutil.rmtree(dbt_repo_dir_name, ignore_errors=True)


@flow(
    name="Transform and Catalog",
    description="Build specified dbt model(s) and upload generated metadata to Luma.",
    timeout_seconds=2 * 60 * 60,
)
def transform_and_catalog(  # noqa: PLR0913
    dbt_repo_url: str | None = None,
    dbt_repo_url_secret: str | None = None,
    dbt_project_path: str = "dbt",
    dbt_repo_branch: str | None = None,
    dbt_repo_token_secret: str | None = None,
    dbt_selects: dict[str, str] | None = None,
    dbt_target: str | None = None,
    dbt_target_dir_path: str | None = None,
    luma_url: str = "http://localhost:8000",
    luma_follow: bool = False,
    metadata_kind: Literal["model", "model_run"] = "model_run",
) -> list[str]:
    """Build specified dbt model(s) and upload the generated metadata to Luma.

    Supports ingesting both model and model run metadata (controlled by the
    `metadata_kind` parameter).

    Note that metadata is still ingested even if the preceding `dbt test` task fails.
    This is done in order to capture test failure metadata in the data catalog.

    Args:
        dbt_repo_url (str, optional): The URL for cloning the dbt repo with relevant
            dbt project. Defaults to None.
        dbt_repo_url_secret (str, optional): Alternatively to above, the secret
            containing `dbt_repo_url`. Defaults to None.
        dbt_project_path (str): Path to the dbt project directory, relative to the
            dbt repository's root. For example, "dbt/my_dbt_project". Defaults to "dbt".
        dbt_repo_branch (str, optional): The branch of the dbt repo to use. Defaults to
            None (default repo branch).
        dbt_repo_token_secret (str, optional): The secret containing the personal access
            token used to clone the dbt repository, in case it's private. Not required
            if token is already included in `dbt_repo_url` (which is NOT recommended).
            Defaults to None.
        dbt_selects (dict, optional): Valid
            [dbt node selection](https://docs.getdbt.com/reference/node-selection/syntax)
            expressions. Valid keys are `run`, `test`, and `source_freshness`. The test
                select expression is taken from run's, as long as run select is
                provided. Defaults to None.
        dbt_target (str): The dbt target to use. If not specified, the default dbt
            target (as specified in `profiles.yaml`) will be used. Defaults to None.
        dbt_target_dir_path (str): The path to your dbt project's target
            directory, which contains dbt artifact JSON files, relative
            to dbt project's root directory. By default,
            `<repo_name>/<dbt_project_path>/target`, since "target" is the default
            name of the directory generated by dbt.
        luma_url (str, optional): The URL of the Luma instance to ingest into.
            Defaults to "http://localhost:8000".
        luma_follow (bool, optional): Whether to follow the ingestion process until it's
            completed (by default, ingestion request is sent without awaiting for the
            response). By default, `False`.
        metadata_kind (Literal["model", "model_run"], optional): The kind of metadata
            to ingest. Defaults to "model_run".

    Returns:
        list[str]: Lines from stdout of the `upload_metadata` task as a list.

    Examples:
        # Build staging models.

        ```python
        import os
        from prefect_viadot.flows import transform_and_catalog

        my_dbt_repo_url = "https://github.com/dbt-labs/jaffle_shop"
        my_luma_url = "http://localhost:8000"

        transform_and_catalog(
            dbt_repo_url=my_dbt_repo_url
            dbt_selects={"run": "staging"}
            luma_url=my_luma_url,
        )
        ```

        Some common `dbt_select` patterns:
        - build a model and all its downstream dependencies: `dbt_select="my_model+"`
        - build all models in a directory: `dbt_select="models/staging"`
    """
    # Clone the dbt project.
    dbt_repo_url = dbt_repo_url or get_credentials(dbt_repo_url_secret)
    clone = clone_repo(
        url=dbt_repo_url,
        checkout_branch=dbt_repo_branch,
        token_secret=dbt_repo_token_secret,
    )

    # Prepare the environment.
    dbt_repo_name = dbt_repo_url.split("/")[-1].replace(".git", "")
    dbt_project_path_full = Path(dbt_repo_name) / dbt_project_path
    dbt_pull_deps_task = dbt_task.with_options(name="dbt_deps")
    pull_dbt_deps = dbt_pull_deps_task(
        project_path=dbt_project_path_full,
        command="deps",
        wait_for=[clone],
    )

    # Run dbt commands.
    dbt_target_option = f"-t {dbt_target}" if dbt_target is not None else ""

    if metadata_kind == "model_run":
        # Produce `run-results.json` artifact for Luma ingestion.
        if dbt_selects:
            run_select = dbt_selects.get("run")
            test_select = dbt_selects.get("test", run_select)

            run_select_safe = f"-s {run_select}" if run_select is not None else ""
            test_select_safe = f"-s {test_select}" if test_select is not None else ""
        else:
            run_select_safe = ""
            test_select_safe = ""

        run_task = dbt_task.with_options(name="dbt_run")
        run = run_task(
            project_path=dbt_project_path_full,
            command=f"run {run_select_safe} {dbt_target_option}",
            wait_for=[pull_dbt_deps],
        )

        test_task = dbt_task.with_options(name="dbt_test")
        test = test_task(
            project_path=dbt_project_path_full,
            command=f"test {test_select_safe} {dbt_target_option}",
            raise_on_failure=False,
            wait_for=[run],
        )
        upload_metadata_upstream_task = test

    else:
        # Produce `catalog.json` and `manifest.json` artifacts for Luma ingestion.
        docs_generate_task = dbt_task.with_options(name="dbt_docs_generate")
        docs = docs_generate_task(
            project_path=dbt_project_path_full,
            command="docs generate",
            wait_for=[pull_dbt_deps],
        )
        upload_metadata_upstream_task = docs

    # Upload metadata to Luma.
    if dbt_target_dir_path is None:
        dbt_target_dir_path = dbt_project_path_full / "target"

    upload_metadata = luma_ingest_task(
        metadata_kind=metadata_kind,
        metadata_dir_path=dbt_target_dir_path,
        luma_url=luma_url,
        follow=luma_follow,
        wait_for=[upload_metadata_upstream_task],
    )

    # Cleanup.
    remove_dbt_repo_dir(
        dbt_repo_name,
        wait_for=[allow_failure(upload_metadata)],
    )

    return upload_metadata



"""Source imports."""

from importlib.util import find_spec

from ._duckdb import DuckDB
from ._trino import Trino
from .cloud_for_customers import CloudForCustomers
from .customer_gauge import CustomerGauge
from .epicor import Epicor
from .exchange_rates import ExchangeRates
from .genesys import Genesys
from .hubspot import Hubspot
from .mediatool import Mediatool
from .mindful import Mindful
from .outlook import Outlook
from .salesforce import Salesforce
from .sftp import Sftp
from .sharepoint import Sharepoint
from .sql_server import SQLServer
from .supermetrics import Supermetrics, SupermetricsCredentials
from .uk_carbon_intensity import UKCarbonIntensity


__all__ = [
    "CloudForCustomers",
    "CustomerGauge",
    "Epicor",
    "ExchangeRates",
    "Genesys",
    "Hubspot",
    "Mediatool",
    "Mindful",
    "Outlook",
    "SQLServer",
    "Salesforce",
    "Sftp",
    "Sharepoint",
    "Supermetrics",
    "SupermetricsCredentials",  # pragma: allowlist-secret
    "Trino",
    "UKCarbonIntensity",
]
if find_spec("adlfs"):
    from viadot.sources.azure_data_lake import AzureDataLake  # noqa: F401

    __all__.extend(["AzureDataLake"])
if find_spec("duckdb"):
    from viadot.sources._duckdb import DuckDB  # noqa: F401

    __all__.extend(["DuckDB"])
if find_spec("redshift_connector"):
    from viadot.sources.redshift_spectrum import RedshiftSpectrum  # noqa: F401

    __all__.extend(["RedshiftSpectrum"])
if find_spec("s3fs"):
    from viadot.sources.s3 import S3  # noqa: F401

    __all__.extend(["S3"])
if find_spec("s3fs"):
    from viadot.sources.minio import MinIO  # noqa: F401

    __all__.extend(["MinIO"])
if find_spec("pyrfc"):
    from viadot.sources.sap_bw import SAPBW  # noqa: F401
    from viadot.sources.sap_rfc import SAPRFC, SAPRFCV2  # noqa: F401

    __all__.extend(["SAPBW", "SAPRFC", "SAPRFCV2"])

if find_spec("pyspark"):
    from viadot.sources.databricks import Databricks  # noqa: F401

    __all__.append("Databricks")



"""'test_mediatool.py'."""

import json
from unittest.mock import MagicMock, patch

import pandas as pd
import pytest

from viadot.exceptions import CredentialError
from viadot.sources import Mediatool
from viadot.sources.mediatool import MediatoolCredentials


variables = {
    "credentials": {"user_id": "test_user", "token": "test_token"},
    "organizations": {
        "organizations": [{"_id": "1", "name": "Org1", "abbreviation": "O1"}]
    },
    "media_entries": {"mediaEntries": [{"_id": "1", "name": "Entry1"}]},
    "vehicle": {"vehicle": {"_id": "1", "name": "Vehicle1"}},
    "campaigns": {"campaigns": [{"_id": "1", "name": "Campaign1"}]},
    "media_types": {"mediaType": {"_id": "1", "name": "Type1", "type": "Type"}},
}


@pytest.mark.basic
def test_mediatool_credentials():
    """Test Mediatool credentials."""
    MediatoolCredentials(user_id="test_user", token="test_token")  # noqa: S106


@pytest.mark.basic
@patch("viadot.sources.mediatool.get_source_credentials", return_value=None)
def test_missing_credentials(mock_get_source_credentials):
    """Test raise error without credentials."""
    with pytest.raises(CredentialError):
        Mediatool()

    mock_get_source_credentials.assert_called_once()


@pytest.mark.functions
@patch("viadot.sources.mediatool.handle_api_response")
def test_get_organizations(mock_handle_api_response):
    """Test Mediatool `_get_organizations` function."""
    mock_response = MagicMock()
    mock_response.text = json.dumps(variables["organizations"])
    mock_handle_api_response.return_value = mock_response

    mediatool = Mediatool(credentials=variables["credentials"])

    result = mediatool._get_organizations(user_id="test_user")
    expected_result = [{"_id": "1", "name": "Org1", "abbreviation": "O1"}]
    assert result == expected_result


@pytest.mark.functions
@patch("viadot.sources.mediatool.handle_api_response")
def test_get_media_entries(mock_handle_api_response):
    """Test Mediatool `_get_media_entries` function."""
    mock_response = MagicMock()
    mock_response.text = json.dumps(variables["media_entries"])
    mock_handle_api_response.return_value = mock_response

    mediatool = Mediatool(credentials=variables["credentials"])

    result = mediatool._get_media_entries(organization_id="org_id")
    expected_result = [{"_id": "1", "name": "Entry1"}]
    assert result == expected_result


@pytest.mark.functions
@patch("viadot.sources.mediatool.handle_api_response")
def test_get_vehicles(mock_handle_api_response):
    """Test Mediatool `_get_vehicles` function."""
    mock_response = MagicMock()
    mock_response.text = json.dumps(variables["vehicle"])
    mock_handle_api_response.return_value = mock_response

    mediatool = Mediatool(credentials=variables["credentials"])

    result = mediatool._get_vehicles(vehicle_ids=["1"])
    expected_result = [{"_id": "1", "name": "Vehicle1"}]
    assert result == expected_result


@pytest.mark.functions
@patch("viadot.sources.mediatool.handle_api_response")
def test_get_campaigns(mock_handle_api_response):
    """Test Mediatool `_get_campaigns` function."""
    mock_response = MagicMock()
    mock_response.text = json.dumps(variables["campaigns"])
    mock_handle_api_response.return_value = mock_response

    mediatool = Mediatool(credentials=variables["credentials"])

    result = mediatool._get_campaigns(organization_id="org_id")
    expected_result = [{"_id": "1", "name": "Campaign1"}]
    assert result == expected_result


@pytest.mark.functions
@patch("viadot.sources.mediatool.handle_api_response")
def test_get_media_types(mock_handle_api_response):
    """Test Mediatool `_get_media_types` function."""
    mock_response = MagicMock()
    mock_response.text = json.dumps(variables["media_types"])
    mock_handle_api_response.return_value = mock_response

    mediatool = Mediatool(credentials=variables["credentials"])

    result = mediatool._get_media_types(media_type_ids=["1"])
    expected_result = [{"_id": "1", "name": "Type1", "type": "Type"}]
    assert result == expected_result


@pytest.mark.functions
def test_rename_columns():
    """Test Mediatool `_rename_columns` function."""
    df = pd.DataFrame({"col1": [1, 2], "col2": [3, 4]})

    mediatool = Mediatool(credentials=variables["credentials"])

    result = mediatool._rename_columns(df, column_suffix="test")
    expected_result = pd.DataFrame({"col1_test": [1, 2], "col2_test": [3, 4]})
    pd.testing.assert_frame_equal(result, expected_result)


@pytest.mark.connect
@patch("viadot.sources.mediatool.handle_api_response")
def test_api_connection(mock_handle_api_response):
    """Test Mediatool `api_connection` method."""
    mock_response = MagicMock()
    mock_response.text = json.dumps(variables["organizations"])
    mock_handle_api_response.return_value = mock_response

    mediatool = Mediatool(credentials=variables["credentials"])

    result = mediatool.api_connection(get_data_from="organizations")
    expected_result = [{"_id": "1", "name": "Org1", "abbreviation": "O1"}]
    assert result == expected_result


@pytest.mark.functions
@patch("viadot.sources.mediatool.handle_api_response")
def test_to_df(mock_handle_api_response):
    """Test Mediatool `to_df` method."""
    mock_response = MagicMock()
    mock_response.text = json.dumps({"mediaEntries": [{"_id": "1", "name": "Entry1"}]})
    mock_handle_api_response.return_value = mock_response

    mediatool = Mediatool(credentials=variables["credentials"])

    data = [{"_id": "1", "name": "Entry1"}]
    result_df = mediatool.to_df(data=data, column_suffix="media_entries")
    result_df.drop(
        columns=["_viadot_source", "_viadot_downloaded_at_utc"],
        inplace=True,
        axis=1,
    )
    expected_result = pd.DataFrame(
        {"_id_media_entries": ["1"], "name_media_entries": ["Entry1"]}
    )
    pd.testing.assert_frame_equal(result_df, expected_result)



from pathlib import Path
from unittest.mock import MagicMock, patch

import pandas as pd
import pytest
import sharepy
from sharepy.errors import AuthError

from viadot.exceptions import CredentialError
from viadot.sources import Sharepoint
from viadot.sources.sharepoint import SharepointCredentials


DUMMY_CREDS = {"site": "test", "username": "test2", "password": "test"}
SAMPLE_DF = pd.DataFrame(
    {
        "int_col": [1, 2, 3, 4, 5, None],
        "float_col": [1.1, 2.2, 3.3, 3.0, 5.5, 6.6],
        "str_col": ["a", "b", "c", "d", "e", "f"],
        "nan_col": [None, None, None, None, None, None],
        "mixed_col": [1, "text", None, None, 4.2, "text2"],
    }
)


class SharepointMock(Sharepoint):
    def get_connection(self):
        return sharepy.session.SharePointSession

    def _download_file_stream(self, url: str | None = None, **kwargs):  # noqa: ARG002
        if "nrows" in kwargs:
            msg = "Parameter 'nrows' is not supported."
            raise ValueError(msg)

        return pd.ExcelFile(Path("tests/unit/test_file.xlsx"))


@pytest.fixture
def sharepoint_mock():
    return SharepointMock(credentials=DUMMY_CREDS)


def test_valid_credentials():
    credentials = {
        "site": "tenant.sharepoint.com",
        "username": "user@example.com",
        "password": "password",
    }
    shrp_creds = SharepointCredentials(**credentials)
    assert shrp_creds.site == credentials["site"]
    assert shrp_creds.username == credentials["username"]
    assert shrp_creds.password == credentials["password"]


def test_invalid_authentication():
    credentials = {
        "site": "tenant.sharepoint.com",
        "username": "user@example.com",
        "password": "password",
    }

    s = Sharepoint(credentials=credentials)

    # Patch the sharepy.connect method to simulate an authentication failure
    with patch("sharepy.connect") as mock_connect:
        mock_connect.side_effect = AuthError("Authentication failed")

        with pytest.raises(
            CredentialError,
            match="Could not authenticate to tenant.sharepoint.com with provided credentials.",
        ):
            s.get_connection()


def test_missing_username():
    credentials = {"site": "example.sharepoint.com", "password": "password"}
    with pytest.raises(
        CredentialError,
        match="'site', 'username', and 'password' credentials are required.",
    ):
        SharepointCredentials(**credentials)


def test_sharepoint_default_na(sharepoint_mock):
    df = sharepoint_mock.to_df(
        url="test/file.xlsx", na_values=Sharepoint.DEFAULT_NA_VALUES
    )

    assert not df.empty
    assert "NA" not in list(df["col_a"])


def test_sharepoint_custom_na(sharepoint_mock):
    df = sharepoint_mock.to_df(
        url="test/file.xlsx",
        na_values=[v for v in Sharepoint.DEFAULT_NA_VALUES if v != "NA"],
    )

    assert not df.empty
    assert "NA" in list(df["col_a"])


def test__get_file_extension(sharepoint_mock):
    url_excel = "https://tenant.sharepoint.com/sites/site/file.xlsx"
    url_dir = "https://tenant.sharepoint.com/sites/site/"
    url_txt = "https://tenant.sharepoint.com/sites/site/file.txt"

    excel_ext = sharepoint_mock._get_file_extension(url=url_excel)
    txt_ext = sharepoint_mock._get_file_extension(url=url_txt)
    dir_ext = sharepoint_mock._get_file_extension(url=url_dir)

    assert excel_ext == ".xlsx"
    assert txt_ext == ".txt"
    assert dir_ext == ""


def test__is_file(sharepoint_mock):
    is_file = sharepoint_mock._is_file(url="https://example.com/file.xlsx")
    assert is_file is True

    is_file = sharepoint_mock._is_file(url="https://example.com/dir")
    assert is_file is False


def test__parse_excel_single_sheet(sharepoint_mock):
    excel_file = sharepoint_mock._download_file_stream()
    result_df = sharepoint_mock._parse_excel(excel_file, sheet_name="Sheet1")
    expected = pd.DataFrame(
        {
            "col_a": ["val1", "", "val2", "NA", "N/A", "#N/A"],
            "col_b": ["val1", "val2", "val3", "val4", "val5", "val6"],
        }
    )

    assert result_df["col_b"].equals(expected["col_b"])


def test__parse_excel_string_dtypes(sharepoint_mock):
    excel_file = sharepoint_mock._download_file_stream()
    result_df = sharepoint_mock._parse_excel(excel_file, sheet_name="Sheet1")

    for column in result_df.columns:
        assert result_df[column].dtype == object


def test__load_and_parse_not_valid_extension(sharepoint_mock):
    with pytest.raises(ValueError):  # noqa: PT011
        sharepoint_mock._load_and_parse(file_url="https://example.com/file.txt")


def test_scan_sharepoint_folder_valid_url(sharepoint_mock):
    url = "https://company.sharepoint.com/sites/site_name/final_folder/"

    # Mock the response from SharePoint
    mock_response = MagicMock()
    mock_response.json.return_value = {
        "d": {
            "results": [
                {"Name": "file1.txt"},
                {"Name": "file2.txt"},
            ]
        }
    }

    # Inject the mock response
    sharepoint_mock.get_connection().get = MagicMock(return_value=mock_response)

    expected_files = [
        "https://company.sharepoint.com/sites/site_name/final_folder/file1.txt",
        "https://company.sharepoint.com/sites/site_name/final_folder/file2.txt",
    ]

    result = sharepoint_mock.scan_sharepoint_folder(url)
    assert result == expected_files


def test_scan_sharepoint_folder_invalid_url(sharepoint_mock):
    url = "https://company.sharepoint.com/folder/sub_folder/final_folder"

    with pytest.raises(ValueError, match="URL does not contain '/sites/' segment."):
        sharepoint_mock.scan_sharepoint_folder(url)


def test_scan_sharepoint_folder_empty_response(sharepoint_mock):
    url = (
        "https://company.sharepoint.com/sites/site_name/folder/sub_folder/final_folder"
    )

    mock_response = MagicMock()
    mock_response.json.return_value = {"d": {"results": []}}

    sharepoint_mock.get_connection().get = MagicMock(return_value=mock_response)

    result = sharepoint_mock.scan_sharepoint_folder(url)
    assert result == []


def test_download_file_stream_unsupported_param(sharepoint_mock):
    url = "https://company.sharepoint.com/sites/site_name/folder/test_file.xlsx"

    with pytest.raises(ValueError, match="Parameter 'nrows' is not supported."):
        sharepoint_mock._download_file_stream(url, nrows=10)



"""Tasks for interacting with Microsoft Sharepoint."""

from typing import Any

import pandas as pd
from prefect import get_run_logger, task

from viadot.orchestration.prefect.exceptions import MissingSourceCredentialsError
from viadot.orchestration.prefect.utils import get_credentials
from viadot.sources import Sharepoint


@task(retries=3, retry_delay_seconds=10, timeout_seconds=60 * 60)
def sharepoint_to_df(
    url: str,
    sheet_name: str | list[str | int] | int | None = None,
    columns: str | list[str] | list[int] | None = None,
    tests: dict[str, Any] | None = None,
    file_sheet_mapping: dict | None = None,
    na_values: list[str] | None = None,
    credentials_secret: str | None = None,
    config_key: str | None = None,
) -> pd.DataFrame:
    """Load an Excel file stored on Microsoft Sharepoint into a pandas `DataFrame`.

    Modes:
    If the `URL` ends with the file (e.g ../file.xlsx) it downloads only the file and
    creates a DataFrame from it.
    If the `URL` ends with the folder (e.g ../folder_name/): it downloads multiple files
    and creates a DataFrame from them:
        - If `file_sheet_mapping` is provided, it downloads and processes only
            the specified files and sheets.
        - If `file_sheet_mapping` is NOT provided, it downloads and processes all of
            the files from the chosen folder.

    Args:
        url (str): The URL to the file.
        sheet_name (str | list | int, optional): Strings are used
            for sheet names. Integers are used in zero-indexed sheet positions
            (chart sheets do not count as a sheet position). Lists of strings/integers
            are used to request multiple sheets. Specify None to get all worksheets.
            Defaults to None.
        columns (str | list[str] | list[int], optional): Which columns to ingest.
            Defaults to None.
        credentials_secret (str, optional): The name of the secret storing
            the credentials. Defaults to None.
            More info on: https://docs.prefect.io/concepts/blocks/
        file_sheet_mapping (dict): A dictionary where keys are filenames and values are
            the sheet names to be loaded from each file. If provided, only these files
            and sheets will be downloaded. Defaults to None.
        na_values (list[str] | None): Additional strings to recognize as NA/NaN.
            If list passed, the specific NA values for each column will be recognized.
            Defaults to None.
        tests (dict[str], optional): A dictionary with optional list of tests
                to verify the output dataframe. If defined, triggers the `validate`
                function from viadot.utils. Defaults to None.
        config_key (str, optional): The key in the viadot config holding relevant
            credentials. Defaults to None.

    Returns:
        pd.Dataframe: The pandas `DataFrame` containing data from the file.
    """
    if not (credentials_secret or config_key):
        raise MissingSourceCredentialsError

    logger = get_run_logger()

    credentials = get_credentials(secret_name=credentials_secret)
    s = Sharepoint(credentials=credentials, config_key=config_key)

    logger.info(f"Downloading data from {url}...")
    df = s.to_df(
        url,
        sheet_name=sheet_name,
        tests=tests,
        usecols=columns,
        na_values=na_values,
        file_sheet_mapping=file_sheet_mapping,
    )
    logger.info(f"Successfully downloaded data from {url}.")

    return df


@task(retries=3, retry_delay_seconds=10, timeout_seconds=60 * 60)
def sharepoint_download_file(
    url: str,
    to_path: str,
    credentials_secret: str | None = None,
    config_key: str | None = None,
) -> None:
    """Download a file from Sharepoint.

    Args:
        url (str): The URL of the file to be downloaded.
        to_path (str): Where to download the file.
        credentials_secret (str, optional): The name of the secret that stores
            Sharepoint credentials. Defaults to None.
        credentials (SharepointCredentials, optional): Sharepoint credentials.
        config_key (str, optional): The key in the viadot config holding relevant
            credentials.
    """
    if not (credentials_secret or config_key):
        raise MissingSourceCredentialsError

    logger = get_run_logger()

    credentials = get_credentials(secret_name=credentials_secret)
    s = Sharepoint(credentials=credentials, config_key=config_key)

    logger.info(f"Downloading data from {url}...")
    s.download_file(url=url, to_path=to_path)
    logger.info(f"Successfully downloaded data from {url}.")



"""'mediatool.py'."""

import json

import pandas as pd
from pydantic import BaseModel

from viadot.config import get_source_credentials
from viadot.exceptions import APIError, CredentialError
from viadot.sources.base import Source
from viadot.utils import add_viadot_metadata_columns, handle_api_response


class MediatoolCredentials(BaseModel):
    """Checking for values in Mediatool credentials dictionary.

    Two key values are held in the Mediatool connector:
        - user_id: The unique ID for the organization.
        token: A unique token to be used as the password for API requests.

    Args:
        BaseModel (pydantic.main.ModelMetaclass): A base class for creating
            Pydantic models.
    """

    user_id: str
    token: str


class Mediatool(Source):
    """Class implementing the Mediatool API.

    Download data from Mediatool platform. Using Mediatool class user is able to
    download organizations, media entries, campaigns, vehicles, and media types data.
    """

    def __init__(
        self,
        *args,
        credentials: MediatoolCredentials | None = None,
        config_key: str | None = None,
        user_id: str | None = None,
        **kwargs,
    ):
        """Create an instance of the Mediatool class.

        Args:
            credentials (MediatoolCredentials, optional): Meditaool credentials.
                Defaults to None.
            config_key (str, optional): The key in the viadot config holding relevant
                credentials. Defaults to None.
            user_id (str, optional): User ID. Defaults to None.
        """
        credentials = credentials or get_source_credentials(config_key) or None
        if credentials is None:
            message = "Missing credentials."
            raise CredentialError(message)

        validated_creds = dict(MediatoolCredentials(**credentials))
        super().__init__(*args, credentials=validated_creds, **kwargs)

        self.header = {"Authorization": f"Bearer {credentials.get('token')}"}
        self.user_id = user_id or credentials.get("user_id")

        self.url_abbreviation = None

    def _rename_columns(
        self,
        df: pd.DataFrame,
        column_suffix: str,
    ) -> pd.DataFrame:
        """Rename columns.

        Args:
            df (pd.DataFrame): Incoming Data frame.
            column_suffix (str): String to be added at the end of column name.

        Returns:
            pd.DataFrame: Modified Data Frame.
        """
        column_suffix = column_suffix.split("get_")[-1]
        dict_mapped_names = {
            column_name: f"{column_name}_{column_suffix}" for column_name in df.columns
        }

        return df.rename(columns=dict_mapped_names)

    def _get_organizations(
        self,
        user_id: str | None = None,
    ) -> list[dict[str, str]]:
        """Get organizations data based on the user ID.

        Args:
            user_id (str, optional): User ID. Defaults to None.

        Returns:
            list[dict[str, str]]: A list of dicts will be returned.
        """
        user_id = user_id or self.user_id
        url_organizations = f"https://api.mediatool.com/users/{user_id}/organizations"

        response = handle_api_response(
            url=url_organizations,
            headers=self.header,
            method="GET",
        )
        response_dict = json.loads(response.text)
        organizations = response_dict["organizations"]

        list_organizations = []
        for org in organizations:
            list_organizations.append(
                {
                    "_id": org.get("_id"),
                    "name": org.get("name"),
                    "abbreviation": org.get("abbreviation"),
                }
            )

        return list_organizations

    def _get_media_entries(
        self,
        organization_id: str,
    ) -> list[dict[str, str]]:
        """Data for media entries.

        Args:
            organization_id (str): Organization ID.

        Returns:
            list[dict[str, str]]: A list of dicts will be returned.
        """
        url = (
            "https://api.mediatool.com/searchmediaentries?q="
            + f'{{"organizationId": "{organization_id}"}}'
        )

        response = handle_api_response(
            url=url,
            headers=self.header,
            method="GET",
        )
        response_dict = json.loads(response.text)

        return response_dict["mediaEntries"]

    def _get_vehicles(
        self,
        vehicle_ids: list[str],
    ) -> list[dict[str, str]]:
        """Vehicles data based on the organization IDs.

        Args:
            vehicle_ids (list[str]): List of organization IDs.

        Raises:
            APIError: Mediatool API does not recognise the vehicle id.

        Returns:
            list[dict[str, str]]: A list of dicts will be returned.
        """
        response_list = []
        missing_vehicles = []

        for vid in vehicle_ids:
            url = f"https://api.mediatool.com/vehicles/{vid}"
            try:
                response = handle_api_response(
                    url=url,
                    headers=self.header,
                    method="GET",
                )
            except APIError:
                missing_vehicles.append(vid)
            else:
                response_dict = json.loads(response.text)
                response_list.append(response_dict["vehicle"])

        if missing_vehicles:
            self.logger.error(f"Vehicle were not found for: {missing_vehicles}.")

        return response_list

    def _get_campaigns(
        self,
        organization_id: str,
    ) -> list[dict[str, str]]:
        """Campaign data based on the organization ID.

        Args:
            organization_id (str): Organization ID.

        Returns:
            list[dict[str, str]]: A list of dicts will be returned.
        """
        url_campaigns = (
            f"https://api.mediatool.com/organizations/{organization_id}/campaigns"
        )

        response = handle_api_response(
            url=url_campaigns,
            headers=self.header,
            method="GET",
        )
        response_dict = json.loads(response.text)

        return response_dict["campaigns"]

    def _get_media_types(
        self,
        media_type_ids: list[str],
    ) -> list[dict[str, str]]:
        """Media types data based on the media types ID.

        Args:
            media_type_ids (list[str]): List of media type IDs.

        Returns:
            list[dict[str, str]]: A list of dicts will be returned.
        """
        list_media_types = []
        for id_media_type in media_type_ids:
            response = handle_api_response(
                url=f"https://api.mediatool.com/mediatypes/{id_media_type}",
                headers=self.header,
                method="GET",
            )
            response_dict = json.loads(response.text)
            list_media_types.append(
                {
                    "_id": response_dict.get("mediaType").get("_id"),
                    "name": response_dict.get("mediaType").get("name"),
                    "type": response_dict.get("mediaType").get("type"),
                }
            )

        return list_media_types

    def api_connection(
        self,
        get_data_from: str,
        organization_id: str | None = None,
        vehicle_ids: list[str] | None = None,
        media_type_ids: list[str] | None = None,
    ) -> list[dict[str, str]]:
        """General method to connect to Mediatool API and generate the response.

        Args:
            get_data_from (str): Method to be used to extract data from.
            organization_id (str, optional): Organization ID. Defaults to None.
            vehicle_ids (list[str]): List of organization IDs. Defaults to None.
            media_type_ids (list[str]): List of media type IDs. Defaults to None.

        Returns:
            list[dict[str, str]]: Data from Mediatool API connection.
        """
        self.url_abbreviation = get_data_from

        if self.url_abbreviation == "organizations":
            returned_data = self._get_organizations(self.user_id)

        elif self.url_abbreviation == "media_entries":
            returned_data = self._get_media_entries(organization_id=organization_id)

        elif self.url_abbreviation == "vehicles":
            returned_data = self._get_vehicles(vehicle_ids=vehicle_ids)

        elif self.url_abbreviation == "campaigns":
            returned_data = self._get_campaigns(organization_id=organization_id)

        elif self.url_abbreviation == "media_types":
            returned_data = self._get_media_types(media_type_ids=media_type_ids)

        return returned_data

    @add_viadot_metadata_columns
    def to_df(
        self,
        if_empty: str = "warn",
        **kwargs,
    ) -> pd.DataFrame:
        """Pandas Data Frame with the data in the Response object and metadata.

        Args:
            if_empty (str, optional): What to do if a fetch produce no data.
                Defaults to "warn

        Returns:
            pd.Dataframe: The response data as a Pandas Data Frame plus viadot metadata.
        """
        data = kwargs.get("data", False)
        column_suffix = kwargs.get("column_suffix", None)
        columns = kwargs.get("columns", None)

        super().to_df(if_empty=if_empty)

        data_frame = pd.DataFrame.from_dict(data)

        if column_suffix == "campaigns":
            data_frame.replace(
                to_replace=[r"\\t|\\n|\\r", "\t|\n|\r"],
                value=["", ""],
                regex=True,
                inplace=True,
            )

        if column_suffix:
            data_frame = self._rename_columns(
                df=data_frame, column_suffix=column_suffix
            )

        if columns:
            if set(columns).issubset(set(data_frame.columns)):
                data_frame = data_frame[columns]
            elif not set(columns).issubset(set(data_frame.columns)):
                self.logger.error(
                    f"Columns '{', '.join(columns)}' are incorrect. "
                    + "Whole dictionary for 'mediaEntries' will be returned."
                )

        if data_frame.empty:
            self._handle_if_empty(
                if_empty=if_empty,
                message="The response does not contain any data.",
            )
        else:
            self.logger.info(
                "Successfully downloaded data from "
                + f"the Mediatool API ({self.url_abbreviation})."
            )

        return data_frame



"""Common utilities for use in tasks."""

import contextlib
import json
from json.decoder import JSONDecodeError
import logging
import os
import sys
import tempfile
from typing import Any

import anyio
from anyio import open_process
from anyio.streams.text import TextReceiveStream
from prefect.blocks.system import Secret
from prefect.client.orchestration import PrefectClient
from prefect.settings import PREFECT_API_KEY, PREFECT_API_URL


with contextlib.suppress(ModuleNotFoundError):
    from prefect_aws import AwsCredentials
    from prefect_aws.secrets_manager import AwsSecret
from prefect_sqlalchemy import DatabaseCredentials

from viadot.orchestration.prefect.exceptions import MissingPrefectBlockError


with contextlib.suppress(ModuleNotFoundError):
    from prefect_azure import AzureKeyVaultSecretReference


async def list_block_documents() -> list[Any]:
    """Retrieve list of Prefect block documents."""
    async with PrefectClient(
        api=PREFECT_API_URL.value(), api_key=PREFECT_API_KEY.value()
    ) as client:
        return await client.read_block_documents()


def _get_azure_credentials(secret_name: str) -> dict[str, Any]:
    """Retrieve credentials from the Prefect 'AzureKeyVaultSecretReference' block.

    Args:
        secret_name (str): The name of the secret to be retrieved.

    Returns:
        dict: A dictionary containing the credentials.
    """
    try:
        credentials = json.loads(
            AzureKeyVaultSecretReference.load(secret_name).get_secret()
        )
    except JSONDecodeError:
        credentials = AzureKeyVaultSecretReference.load(secret_name).get_secret()

    return credentials


def _get_aws_credentials(
    secret_name: str,
    block_type: str,
) -> dict[str, Any] | str:
    """Retrieve credentials from the Prefect 'AwsSecret' block document.

        It distinguishes the types of block to take a correct action.
        This is needed as AwsSecret block stores credentials that can be used in tasks
        which ingest data from different systems (for ex. SAP), where
        AwsCredentials block stores AWS credentials solely and this could
        be needed in other tasks which connects
        to AWS directly ( for ex. redshift spectrum , s3).

    Args:
        secret_name (str): The name of the secret to be retrieved.
        block_type (str): Type of prefect block

    Returns:
        dict | str: A dictionary or a string containing the credentials.
    """
    if block_type == "AwsSecret":
        aws_secret_block = AwsSecret.load(secret_name)
        secret = aws_secret_block.read_secret()
        try:
            credentials = json.loads(secret)
        except json.JSONDecodeError:
            credentials = secret
    elif block_type == "AwsCredentials":
        aws_credentials_block = AwsCredentials.load(secret_name)
        credentials = {
            "aws_access_key_id": aws_credentials_block.aws_access_key_id,
            "aws_secret_access_key": aws_credentials_block.aws_secret_access_key.get_secret_value(),
            "region_name": aws_credentials_block.region_name,
        }

    return credentials


def _get_secret_credentials(secret_name: str) -> dict[str, Any] | str:
    """Retrieve credentials from the Prefect 'Secret' block document.

    Args:
        secret_name (str): The name of the secret to be retrieved.

    Returns:
        dict | str: A dictionary or a string containing the credentials.
    """
    secret = Secret.load(secret_name).get()
    try:
        credentials = json.loads(secret)
    except json.JSONDecodeError:
        credentials = secret

    return credentials


def _get_database_credentials(secret_name: str) -> dict[str, Any] | str:
    """Retrieve credentials from the Prefect 'DatabaseCredentials' block document.

    Args:
        secret_name (str): The name of the secret to be retrieved.

    Returns:
        dict | str: A dictionary or a string containing the credentials.
    """
    secret = DatabaseCredentials.load(name=secret_name).dict()

    credentials = secret
    credentials["user"] = secret.get("username")
    credentials["db_name"] = secret.get("database")
    credentials["password"] = secret.get("password").get_secret_value()
    if secret.get("port"):
        credentials["server"] = secret.get("host") + "," + str(secret.get("port"))
    else:
        credentials["server"] = secret.get("host")

    return credentials


def get_credentials(secret_name: str) -> dict[str, Any]:
    """Retrieve credentials from the Prefect block document.

    Args:
        secret_name (str): The name of the secret to be retrieved.

    Returns:
        dict: A dictionary containing the credentials.
    """
    # Prefect does not allow upper case letters for blocks,
    # so some names might be lowercased versions of the original

    secret_name_lowercase = secret_name.lower()
    blocks = anyio.run(list_block_documents)

    for block in blocks:
        if block.name == secret_name_lowercase:
            block_type = block.block_schema.fields["title"]
            break
    else:
        msg = "The provided secret name is not valid."
        raise MissingPrefectBlockError(msg)

    if block_type in ("AwsSecret", "AwsCredentials"):
        credentials = _get_aws_credentials(secret_name, block_type)
    elif block_type == "AzureKeyVaultSecretReference":
        credentials = _get_azure_credentials(secret_name)
    elif block_type == "DatabaseCredentials":
        credentials = _get_database_credentials(secret_name)
    elif block_type == "Secret":
        credentials = _get_secret_credentials(secret_name)
    else:
        msg = f"The provided secret block type: {block_type} is not supported"
        raise MissingPrefectBlockError(msg)

    return credentials


async def shell_run_command(
    command: str,
    env: dict[str, Any] | None = None,
    helper_command: str | None = None,
    shell: str = "bash",
    return_all: bool = False,
    stream_level: int = logging.INFO,
    logger: logging.Logger | None = None,
    raise_on_failure: bool = True,
) -> list[str] | str:
    """Runs arbitrary shell commands as a util.

    Args:
        command: Shell command to be executed; can also be
            provided post-initialization by calling this task instance.
        env: Dictionary of environment variables to use for
            the subprocess; can also be provided at runtime.
        helper_command: String representing a shell command, which
            will be executed prior to the `command` in the same process.
            Can be used to change directories, define helper functions, etc.
            for different commands in a flow.
        shell: Shell to run the command with; defaults to "bash".
        return_all: Whether this task should return all lines of stdout as a list,
            or just the last line as a string; defaults to `False`.
        stream_level: The logging level of the stream.
        logger: Can pass a desired logger; if not passed, will automatically
            gets a run logger from Prefect.
        raise_on_failure: Whether to raise an exception if the command fails.

    Returns:
        If return all, returns all lines as a list; else the last line as a string.

    Example:
        Echo "hey it works".
        ```python
        from prefect_shell.utils import shell_run_command
        await shell_run_command("echo hey it works")
        ```
    """
    if logger is None:
        logging.basicConfig(level=logging.INFO)
        logger = logging.getLogger("prefect_shell.utils")

    current_env = os.environ.copy()
    current_env.update(env or {})

    with tempfile.NamedTemporaryFile(prefix="prefect-") as tmp:
        if helper_command:
            tmp.write(helper_command.encode())
            tmp.write(os.linesep.encode())
        tmp.write(command.encode())
        tmp.flush()

        shell_command = [shell, tmp.name]
        if sys.platform == "win32":
            shell_command = " ".join(shell_command)

        lines = []
        async with await open_process(shell_command, env=env) as process:
            async for text in TextReceiveStream(process.stdout):
                logger.log(stream_level, text)
                lines.extend(text.rstrip().split("\n"))

            await process.wait()
            if process.returncode:
                stderr = "\n".join(
                    [text async for text in TextReceiveStream(process.stderr)]
                )
                if not stderr and lines:
                    stderr = f"{lines[-1]}\n"
                msg = (
                    f"Command failed with exit code {process.returncode}:\n" f"{stderr}"
                )
                if raise_on_failure:
                    raise RuntimeError(msg)
                lines.append(msg)

    return lines if return_all else lines[-1]
